#
# RevisionManager
#
# Copyright (c) 2001 by beehive elektronische medien GmbH
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: RevisionManager.py,v 1.32 2003/02/19 08:27:22 philipp Exp $
"""

import os
from StringIO import StringIO

import Globals
from OFS import ObjectManager, SimpleItem, Image
from OFS.FindSupport import FindSupport
from AccessControl import Role, ClassSecurityInfo
from App.Dialogs import MessageDialog
from Acquisition import aq_parent, aq_base, aq_acquire
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZPublisher.mapply import mapply

import Permissions
from CVSWrapper import CVSWrapper, CVSError
from FactoryRegistry import FactoryRegistry
from IRevisionManager import IRevisionManager
from IRevisionManagerZMI import IRevisionManagerZMI

# constants and global variables
FS_DIRNAME = "RevisionManager"
DATA_DIR = os.path.join(Globals.INSTANCE_HOME, "var", FS_DIRNAME)
CVSPASSFILE = os.path.join(DATA_DIR, 'cvspass')

class RevisionManager(SimpleItem.SimpleItem):
    """
    RevisionManager
    """

    __implements__ = IRevisionManager, IRevisionManagerZMI

    # Zope specific stuff
    manage_options = (
        ( {'label': 'Status', 'action': 'rm_statusForm',
           'help': ('RevisionManager', 'status.stx') },
          {'label': 'Add/Remove', 'action': 'rm_addRemoveForm',
           'help': ('RevisionManager', 'addRemove.stx') },
          {'label': 'Map files to objects', 'action': 'rm_mapFSObjectsForm',
           'help': ('RevisionManager', 'mapFSObjects.stx') },
          {'label': 'Configuration', 'action': 'rm_configForm',
           'help': ('RevisionManager', 'config.stx')},
        )
        + SimpleItem.Item.manage_options
        + Role.RoleManager.manage_options
   )

    id = title = "RevisionManager"
    meta_type = "Revision Manager"
    bad_metatypes = ("User Folder", "Product")
    bad_ids = ("acl_users", "RevisionManager", "Control_Panel")
    security = ClassSecurityInfo()

    def __init__(self):
        self._file_paths = ()
        self._settings = {}

        # create the data directory if it doesn't exist already
        if not os.path.isdir(DATA_DIR):
            os.mkdir(DATA_DIR)

    def _clean_up(self):
        if not self._settings.has_key('module'):
            return
	module = self._settings['module']
	module_path = os.path.join(DATA_DIR, module)
	if os.path.isdir(module_path):
	    try:
	        deltree(module_path)
	    except OSError:
		pass

    def _setId(self, id):
        # Don't permit rename
        if id != self.id:
            return MessageDialog(title="Invalid ID",
                                 message="It is not possible to change the ID of a RevisionManager object.",
                                 action="manage_workspace")

    def _canCopy(self, op=0):
        # Don't permit copying/moving
        return 0
    
    def _generate_wrapper(self, cvs_bin, access_method, repository_path, module,
                          server_addr, user_name, passwd, cvs_rsh, comp_level, debug):
        if hasattr(self, "_cvs_wrapper"):
            del self._cvs_wrapper          # explicitly delete old wrapper
        self._cvs_wrapper = CVSWrapper(
            cvs_bin=cvs_bin,
            access_method=access_method,
            repository_path=repository_path,
            module=module,
            work_dir=DATA_DIR,
            server_addr=server_addr,
            user_name=user_name,
            passwd=passwd,
            cvs_rsh=cvs_rsh,
            comp_level=comp_level,
            cvs_passfile=CVSPASSFILE,
            debug=debug
            )

    def _has_source_methods(self, obj):
        """
        Check whether an object supports FTP/DAV protocols
        """
        base = aq_base(obj)
        # object needs to support at least one method to retrieve its raw source
        # and one method to upload text to it
        return ((hasattr(base, 'manage_FTPget') or hasattr(base, 'EditableBody')
              or hasattr(base, 'document_src')  or hasattr(base, 'read'))
            and (hasattr(base, 'manage_upload') or hasattr(base, 'PUT')))

    def _get_source(self, obj):
        """
        Get the source of an object
        """
        if (hasattr(aq_base(obj), 'data') and hasattr(obj.data, '__class__')
            and obj.data.__class__ is Image.Pdata):
            # yac! we have a File instance with chunked data, let's stream it!
            result = ""
            data = obj.data
            while data is not None:
                result += data.data
                data = data.next
            return result

        #XXX I don't like acquiring the REQUEST, but for some reason, making
        # your own one will screw up the _REAL_ response.
        request = aq_acquire(self, "REQUEST")
        fake_response = HTTPResponse(stdout=StringIO())
        real_response = request.RESPONSE
        request.RESPONSE = fake_response
        kw = { 'REQUEST': request, 'RESPONSE': fake_response }
        
        if hasattr(obj, 'manage_FTPget'):
            result = mapply(obj.manage_FTPget, keyword=kw)
        elif hasattr(obj, 'EditableBody'):
            result = obj.EditableBody()
        elif hasattr(obj, 'document_src'):
            result = mapply(obj.document_src, keyword=kw)
        elif hasattr(obj, 'read'):
            result = obj.read()
        else:
            result = ""

        request.RESPONSE = real_response
        return str(result)
        
    def _put_source(self, obj, file=""):
        """
        Upload text to an object
        """
        if type(file) is not type(""):
            file = file.read()

        base = aq_base(obj)
        if hasattr(base, "manage_upload"):
            obj.manage_upload(file)
        elif hasattr(base, "PUT"):
            # try to acquire the REQUEST
            request = aq_acquire(self, "REQUEST")
            response = HTTPResponse(stdout=StringIO())
            kw = { 'REQUEST': request, 'RESPONSE': response }
            mapply(obj.PUT, keyword=kw)
        else:
            return 1  # error
        return 0

    def _escape_spaces(self, paths):
        escaped = []
        for path in paths:
            escaped.append(path.replace(' ', '\ '))
        return escaped

    ############################################################################
    # Implementation methods for interface
    # IRevisionManager.IRevisionManager
    #

    security.declarePublic("getSettings")
    def getSettings(self):
        dict = {}
        dict.update(self._settings)
        return dict

    security.declareProtected(Permissions.RMManage, "saveSettings")
    def setSettings(self, settings):
        self._settings.update(settings)

    security.declareProtected(Permissions.RMManage, "saveSettings")
    def saveSettings(self, mapping):
        settings = {
            'cvs_bin':         mapping.get("cvs_bin",         "cvs"),
            'access_method':   mapping.get("access_method",   "local"),
            'server_addr':     mapping.get("server_addr",     ""),
            'user_name':       mapping.get("user_name",       ""),
            'passwd':          mapping.get("passwd",          ""),
            'cvs_rsh':         mapping.get("cvs_rsh",         "rsh"),
            'repository_path': mapping.get("repository_path", "/var/cvs"),
            'module':          mapping.get("module",          ""),
            'comp_level':      mapping.get("comp_level",      "3"),
            'debug':           mapping.get("debug",           0)
            }
        self._clean_up()
        self.setSettings(settings)
        self._generate_wrapper(**settings)
        cvs = self._cvs_wrapper
        try:
            cvs.checkout()
        except CVSError:
            pass

    security.declareProtected(Permissions.RMManage, "getFilePaths")
    def getFilePaths(self):
        return self._file_paths

    security.declareProtected(Permissions.RMManage, "getFSFilePaths")
    def getFSFilePaths(self, path=""):
        returnlist = []
        module_path = os.path.join(DATA_DIR, self._settings['module'])
        for filename in os.listdir(os.path.join(module_path, path)):
            filepath = os.path.join(path, filename)
            fullpath = os.path.join(module_path, path, filename)
            if os.path.isfile(fullpath):
                returnlist.append(filepath)
            elif os.path.isdir(fullpath) and filename != "CVS":
                returnlist += self.getFSFilePaths(filepath)
        return tuple(returnlist)

    security.declareProtected(Permissions.RMManage, "getFileInfo")
    def getFileInfo(self, paths=[]):
        return_items = []
        if not len(paths):
            paths = self.getFilePaths()

        # add memorized objects to return list and check their cvs status
        for path in paths:
            try:
                obj = self.restrictedTraverse(path)
                dict = {'path':         path,
                        'meta_type':    obj.meta_type,
                        'absolute_url': obj.absolute_url(),
                        'icon':         obj.icon,
                        'in_zodb':      1 }
            except (AttributeError, KeyError):
                # object has probably been deleted
                dict = {'path':         path,
                        'meta_type':    None,
                        'absolute_url': None,
                        'icon':         None,
                        'in_zodb':      0 }
            return_items.append(dict)
        return tuple(return_items) 

    security.declareProtected(Permissions.RMManage, "getFSFileInfo")
    def getFSFileInfo(self):
        return_items = []
        for path in self.getFSFilePaths():
            try:
                obj = self.restrictedTraverse(path)
            except (AttributeError, KeyError):
                obj = None
                
            in_zodb = obj is not None
            in_items = path in self._file_paths
            
            if not in_zodb:
                full_path = os.path.join(DATA_DIR, self._settings['module'], path)
                path_contents = open(full_path, "rb").read()
                meta_type = guess_meta_type(path_contents)
            else:
                meta_type = obj.meta_type
                
            return_items.append(
                { 'path':      path,
                  'meta_type': meta_type,
                  'in_zodb':   in_zodb,
                  'in_items':  in_items
                  }
                )            
        return tuple(return_items)

    security.declareProtected(Permissions.RMManage, "getSupportedObjectsInfo")
    def getSupportedObjectsInfo(self):
        parent = aq_parent(self)
        finder = FindSupport()
        find_result = finder.ZopeFind(parent, search_sub=1)

        file_paths = list(self._file_paths)
        return_items = []
        for path, obj in find_result:
            if (obj.getId() not in self.bad_ids and           # no bad ids
                obj.meta_type not in self.bad_metatypes):     # no bad meta types
                # if it supports source methods, add it to return list
                if self._has_source_methods(obj):
                    return_items.append(
                        {'path':      path,
                         'icon':      obj.icon,
                         'meta_type': obj.meta_type,
                         'in_items':  path in self._file_paths }
                        )
                    if path in file_paths:
                        file_paths.remove(path)
        # the remaining paths belong to objects which are not in the ZODB anymore
        for path in file_paths:
            return_items.append(
                {'path':      path,
                 'icon':      None,
                 'meta_type': None,
                 'in_items':  1 }
                )
        return tuple(return_items)

    security.declareProtected(Permissions.RMManage, "getSupportedMetaTypes")
    def getSupportedFactoryMetaTypes(self):
        meta_types = FactoryRegistry.getSupportedMetaTypes()
        meta_types = list(meta_types)
        meta_types.sort()
        return tuple(meta_types)

    security.declareProtected(Permissions.RMManage, "getFileStatus")
    def getFileStatus(self, paths=[]):
        if not len(paths):
            self.writeToFS(self.getFilePaths())
        else:
            self.writeToFS(paths)

        paths = self._escape_spaces(paths)
        # get the information from cvs
        cvs = self._cvs_wrapper
        cvs_status = cvs.status(filename=" ".join(paths))

        # first, get the normal file info
        # then, update the dictionaries in-place
        file_info = self.getFileInfo(paths)
        for item in file_info:
            path = item['path']
            if cvs_status.has_key(path):
                dict = {'cvs_status': cvs_status[path]['status'],
                        'cvs_rrev':   cvs_status[path]['rrev'],
                        'cvs_wrev':   cvs_status[path]['wrev'],
                        'cvs_stag':   cvs_status[path]['stag'],
                        'cvs_sopt':   cvs_status[path]['sopt'] }
                del cvs_status[path]
            else:
                dict = {'cvs_status': "Not Available",
                        'cvs_rrev':   "0",
                        'cvs_wrev':   "0",
                        'cvs_stag':   "",
                        'cvs_sopt':   "" }
            item.update(dict)
        return file_info

    security.declareProtected(Permissions.RMManage, "getFileLog")
    def getFileLog(self, paths=[]):
        file_info = self.getFileInfo(paths)

        paths = self._escape_spaces(paths)
        # get the information from cvs
        cvs = self._cvs_wrapper
        cvs_log = cvs.log(filename=" ".join(paths))

        # update dictionaries in-place
        for info in file_info:
            path = info['path']
            if cvs_log.has_key(path):
                info.update(cvs_log[path])
            else:
                info.update({
                    'head':   "N/A",
                    'branch': "N/A",
                    'locks':  "N/A",
                    'access': "N/A",
                    'tags':      {},
                    'revisions': {} 
                    })
        return file_info

    security.declareProtected(Permissions.RMManage, "getModifiedFiles")
    def getModifiedFiles(self, paths=[]):
        modified_files = []
        for status in self.getFileStatus(paths):
            if status['cvs_status'] == "Locally Modified":
                modified_files.append(status['path'])
        return tuple(modified_files)

    security.declareProtected(Permissions.RMUpdate, "update")
    def update(self, paths=[]):
        # write objects to FS
        if len(paths) == 0: 
            self.writeToFS(self._file_paths)
        else:
            self.writeToFS(paths)

        escaped_paths = self._escape_spaces(paths)
        # do the CVS update
        cvs = self._cvs_wrapper
        cvs_output = cvs.update(filename=" ".join(escaped_paths), opt_param="-d")
        # read the files into the objects again
        error_paths = self.readFromFS(paths)
        
        return_list = []
        for path, status in cvs_output.items():
            dict = {'path':     path,
                    'status':   status,
                    'in_zodb':  not(path in error_paths),
                    'in_items': path in self._file_paths }
            return_list.append(dict)
        return tuple(return_list)

    security.declareProtected(Permissions.RMManage, 'getDiff')
    def getDiff(self, path, rev1='', rev2=''):
        path = path.replace(' ', '\ ')
        cvs = self._cvs_wrapper
        diff = cvs.diff(path, rev1, rev2)
        return diff

    security.declareProtected(Permissions.RMManage, 'addFSFileObject')
    def addFSFileObject(self, path, meta_type):
        segments = path.split("/")
        # first make sure that the directory part of the path exists
        # if it doesn't, create as many folders as necessary
        folder = aq_parent(self)
        for id in segments[:-1]:
            subfolder = getattr(folder, id, None)
            if subfolder is None:
                folder.manage_addFolder(id)
                subfolder = getattr(folder, id)
            folder = subfolder

        # now add the object
        # first the the appropriate factory
        factory = FactoryRegistry.getFactory(meta_type)
        id = segments[-1]
        file_name = os.path.join(DATA_DIR, self._settings['module'], path)
        # read the file contents
        data = open(file_name, "rb").read()
        obj = factory.createObject(id, id, data)
        folder._setObject(id, obj)
        return getattr(folder, id)
    
    security.declarePrivate("writeToFS")
    def writeToFS(self, paths, module_path=None):
        if module_path is None:
            module_path = os.path.join(DATA_DIR, self._settings['module'])

        # create module directory if not present already
        if self._settings['module'] not in os.listdir(DATA_DIR):
            try:
                os.mkdir(module_path)
            except OSError:
                pass

        error_paths = []
        for path in paths:
            try:
                obj = self.restrictedTraverse(path)
            except (AttributeError, KeyError):
                error_paths.append(path)
                continue
            fname = os.path.join(module_path, path)
            dirs = path.split("/")

            if not obj.isPrincipiaFolderish:
                # create sub-directories if necessary
                current = module_path
                for dir in dirs[:-1]:
                    if dir not in os.listdir(current):
                        os.mkdir(os.path.join(current, dir))
                    current = os.path.join(current, dir)

                src = self._get_source(obj)
                already_fs = os.path.isfile(fname)
                # check whether the contents has changed
                if already_fs:
                    fp = open(fname, "rb")
                    fs_src = fp.read()
                    fp.close()

                # only write if contents has changed
                if not already_fs or (already_fs and src != fs_src):
                    # open, write to file and close it again
                    fp = open(fname, "wb")
                    fp.write(src)
                    fp.close()
        return error_paths

    security.declarePrivate("readFromFS")
    def readFromFS(self, paths):
        module_path = os.path.join(DATA_DIR, self._settings['module'])
        error_paths = []
        for path in paths:
            try:
                obj = self.restrictedTraverse(path)
            except (AttributeError, KeyError):
                error_paths.append(path)
                continue
            fname = os.path.join(module_path, path)

            if not obj.isPrincipiaFolderish:
                fp = open(fname, "rb")
                src = fp.read()
                fp.close()
                # only modifiy if necessary
                if self._get_source(obj) != src:
                    self._put_source(obj, src)
        return error_paths

    #
    ############################################################################
    # Overwritten SimpleItem methods
    #

    def manage_beforeDelete(self, item, container):
        self._clean_up()

    ############################################################################
    # Implementation methods for interface
    # IRevisionManagerZMI.IRevisionManagerZMI

    # management views
    security.declareProtected(Permissions.RMManage,
                              "rm_statusForm", "rm_addRemoveForm", "rm_configForm",
                              "rm_importForm", "rm_mapFSObjectsForm", "rm_detailsForm",
                              "rm_logForm", "rm_diffForm", "rm_commitForm",
                              "rm_updateForm")
    rm_statusForm       = PageTemplateFile("www/status.pt", globals())
    rm_addRemoveForm    = PageTemplateFile("www/addremove.pt", globals())
    rm_configForm       = PageTemplateFile("www/config.pt", globals())
    rm_importForm       = PageTemplateFile("www/import.pt", globals())
    rm_mapFSObjectsForm = PageTemplateFile("www/mapFSObjects.pt", globals())
    rm_detailsForm      = PageTemplateFile("www/details.pt", globals())
    rm_logForm          = PageTemplateFile("www/log.pt", globals())
    rm_diffForm         = PageTemplateFile("www/diff.pt", globals())
    rm_commitForm       = PageTemplateFile("www/commit.pt", globals())
    rm_updateForm       = PageTemplateFile("www/update.pt", globals())

    security.declareProtected(Permissions.RMAddRemove, "rm_import")
    def rm_import(self, vendor_tag, release_tag, log_message, paths=[], REQUEST=None):
        """
        Properly import selected files
        """
        module = self._settings['module']
        
        vendor_tag = vendor_tag or module.upper()
        release_tag = vendor_tag or module.upper() + "1"
        log_message = log_message or "RevisionManager initial import"
        if vendor_tag == release_tag:
            release_tag += "1"
        
        self._file_paths = tuple(paths)
        try:
            import_dir = os.tmpnam()
        except:
            import_dir = "/tmp/RevisionManager-" + module
        os.mkdir(import_dir)
        self.writeToFS(paths, import_dir)

        # do a CVS import
        cvs = self._cvs_wrapper
        cvs.import_module(import_dir, vendor_tag, release_tag, log_message)
        deltree(import_dir)
        # do a checkout right away as well
        cvs.checkout()

        if REQUEST is not None:
            message = "Module '%s' imported" % module
            return self.rm_addRemoveForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMAddRemove, "rm_mapFSObjects")
    def rm_mapFSObjects(self, items=[], REQUEST=None):
        """
        Map file system files to objects.
        """
        paths = list(self._file_paths)
        for item in items:
            if item.get('add', 0):
                path = item['path']
                if path not in paths:
                    paths.append(path)
                if not item.get('in_zodb', 0):
                    self.addFSFileObject(path, item['meta_type'])
        self._file_paths = tuple(paths)
        if REQUEST is not None:
            return self.rm_statusForm(self, REQUEST)

    security.declareProtected(Permissions.RMConfig, "rm_config")
    def rm_config(self, REQUEST):
        """
        Save configuration settings.
        """
        # only take 'module' from REQUEST if settings doesn't already have it
        module = self._settings.get("module", REQUEST.get("module"))
        REQUEST.set("module", module)
        self.saveSettings(REQUEST)
        if REQUEST is not None:
            message = "Saved changes"
            return self.rm_configForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMAddRemove, "rm_addObjects")
    def rm_addObjects(self, new_paths=[], REQUEST=None, RESPONSE=None):
        """
        Add objects to the repository.
        """
        if not new_paths:
            if RESPONSE is not None:
                RESPONSE.redirect(self.absolute_url() + '/rm_addRemoveForm')
            return

        self.writeToFS(new_paths)

        escaped_new_paths = []
        binary_paths = []
        dir_paths = []
        for path in new_paths:
            # make sure that subdirectories will be added, too
            # do not actually test whether they have a 'CVS' directory already
            # cvs will find out for us much faster than we could do it here
            segments = path.split("/")
            for i in range(1, len(segments)):
                dir_path = "/".join(segments[:i])
                # escape spaces
                dir_path = dir_path.replace(' ', '\ ')
                if dir_path not in dir_paths:
                    dir_paths.append(dir_path)
                    
            # check for binariness
            filename = os.path.join(DATA_DIR, self._settings['module'], path)
            bin = is_binary(open(filename))
            
            # escape spaces
            path = path.replace(' ', '\ ')
            if bin:
                binary_paths.append(path)
            else:
                escaped_new_paths.append(path)

        escaped_new_paths = " ".join(escaped_new_paths)
        binary_paths = " ".join(binary_paths)
        cvs = self._cvs_wrapper
        # first add the directories. They don't need a commit.
        cvs.add(" ".join(dir_paths))
        # now add regular and binary files and commit them.
        if escaped_new_paths:
            cvs.add(escaped_new_paths)
        if binary_paths:
            cvs.add(binary_paths, binary=1)
        cvs.commit(filename=escaped_new_paths + " " + binary_paths,
                   log_message="File added to repository by RevisionManager.")

        # add them to our internal list
        self._file_paths += tuple(new_paths)
        self._p_changed = 1

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_addRemoveForm')
            
    security.declareProtected(Permissions.RMAddRemove, "rm_removeObjects")
    def rm_removeObjects(self, paths=[], REQUEST=None, RESPONSE=None):
        """
        Remove objects from RevisionManager's list and delete their
        equivalents in the CVS repository
        """
        if not paths:
            if RESPONSE is not None:
                RESPONSE.redirect(self.absolute_url() + '/rm_addRemoveForm')
            return

        file_paths = list(self._file_paths)
        escaped_paths = []
        for path in paths:
            # remove from interal list
            file_paths.remove(path)
            # remove file
            filename = os.path.join(DATA_DIR, self._settings['module'], path)
            try:
                if os.path.isdir(filename):
                    os.rmdir(filename)
                else:
                    os.unlink(filename)
            except OSError:
                pass
            # escape spaces
            path = path.replace(' ', '\ ')
            escaped_paths.append(path)

        escaped_paths = " ".join(escaped_paths)
        cvs = self._cvs_wrapper
        cvs.remove(escaped_paths)

	# file won't be removed from the repository unless changes have been
	# committed
        cvs.commit(filename=escaped_paths,
                   log_message="Removed file by RevisionManager.")

        self._file_paths = tuple(file_paths)
 
        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_addRemoveForm')

    security.declareProtected(Permissions.RMCommit, "rm_commit")
    def rm_commit(self, paths=[], log_message="commit", REQUEST=None):
        """
        Commit changes
        """
        if len(paths) == 0:
            self.writeToFS(self._file_paths)
        else:
            self.writeToFS(paths)

        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        output = cvs.commit(filename=" ".join(paths), log_message=log_message)

        if REQUEST is not None:
            message = "Committed files: " + ", ".join(output.keys())
            return self.rm_statusForm(self, REQUEST, manage_tabs_message=message)

    security.declareProtected(Permissions.RMCommit, "rm_setTag")
    def rm_setTag(self, tag, paths=[], REQUEST=None, RESPONSE=None):
        """
        Set tag on file(s)
        """
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        cvs.tag(tag, " ".join(paths))

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMCommit, "rm_setRevision")
    def rm_setRevision(self, setrevision, paths=[], REQUEST=None, RESPONSE=None):
        """
        Set a file's revision manually to setrevision
        """
        paths = self._escape_spaces(paths)
        cvs = self._cvs_wrapper
        cvs.commit(filename=" ".join(paths), opt_param="-r "+setrevision)

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

    security.declareProtected(Permissions.RMUpdate, "rm_revertRevision")
    def rm_revertRevision(self, path, revision, REQUEST=None, RESPONSE=None):
        """
        overwrite an object's data with its data from an older revision
        """
        path = path.replace(' ', '\ ')
        cmdparam = "-p -r" + revision
        cmdtail = ">" + path

        cvs = self._cvs_wrapper
        output = cvs.update(filename=path, opt_param=cmdparam, opt_tail=cmdtail)
        self.readFromFS([path,])

        if RESPONSE is not None:
            RESPONSE.redirect(self.absolute_url() + '/rm_statusForm')

#
# object constructors
#

Globals.InitializeClass(RevisionManager)

def manage_addRevisionManager(self, REQUEST=None, RESPONSE=None):
    """
    Create an instance of RevisionManager
    """
    # prevent instances in folders whose parents already contain an instance
    # (since it may lead to data inconsistency)
    try:
        rm = self.aq_acquire("RevisionManager")
    except AttributeError:
        rm = None
    if rm is not None:
        return MessageDialog(
	    title="Detected a RevisionManager instance",
	    message="There already is a RevisionManager instance in this folder or one of its parent folders.",
	    action="manage_workspace"
	   )

    obj = RevisionManager()
    self.Destination()._setObject(obj.id, obj)
    obj.saveSettings(REQUEST)

    if RESPONSE is not None:
        module_exists = REQUEST.get('module_exists', 0)
        if module_exists:
            method = "rm_mapFSObjectsForm"
        else:
            method = "rm_importForm"
        RESPONSE.redirect(self.DestinationURL() + "/" + obj.getId() + "/" + method)

manage_addForm = PageTemplateFile("www/add.pt", globals())

#
# functions
#

def guess_meta_type(data):
    """
    Guess the meta type of 'data'
    """
    # This is very simple for now
    # If a token in the table is found, it is assummed that the corresponding
    # meta_type is correct
    
    #XXX missing images for now
    table = {
        '<dtml-':            'DTML Method',
        ' tal:':             'Page Template',
        ' Script (Python) ': 'Script (Python)',
        '<params>':          'Z SQL Method'
        }
    default_meta_type = "File"

    for chunk, meta_type in table.items():
        if data.find(chunk) >= 0:
            return meta_type
    return default_meta_type

def is_binary(file):
    """
    Check whether a file represented by file object 'file' contains binary
    data or text
    """
    chunk_size = 4*1024
    # read the file chunk-wise
    data = file.read(chunk_size)
    while data:
        for char in data:
            # binary files usually contain bytes < ASCII 7
            if char <= '\7':
                file.seek(0)
                return 1
        data = file.read(chunk_size)
    return 0

def deltree(dirname):
    """
    works like 'rm -r <dirname>'
    """
    for filename in os.listdir(dirname):
        if os.path.isdir(os.path.join(dirname, filename)):
            deltree(os.path.join(dirname, filename))
        else:
            os.unlink(os.path.join(dirname, filename))
    os.rmdir(dirname)
